// From http://www.redblobgames.com/grids/hexagons/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>
//
// C++11 output

using Ast;
import Output.*;

class OutputCpp extends Output {
    function new() {}

    
    override function Type(t:TExpr) {
        return switch (t) {
            case TVoid: 'void';
            case TInt: 'int';
            case TFloat: 'double';
            case TString: 'const char*';
            case TArray(p): 'vector<${Type(p)}>';
            case _: super.Type(t);
        };
    }

    
    override function Expr(e:AExpr) {
        return switch (e) {
            case AName(name):
                switch (name) {
                    case "PI": "M_PI";
                    case _: '$name';
                }
            case ANew(TArray(_), arguments): '{${[for (a in arguments) Expr(a)].join(", ")}}'; // C++11 initializer
            case ANew(type, arguments): '${Type(type)}(${[for (a in arguments) Expr(a)].join(", ")})';
            case AFunctionCall("#negate", [expr]): '-${Expr(expr)}';
            case AFunctionCall("#not", [expr]): '!${Expr(expr)}';
            case AFunctionCall("$error", [expr]): 'throw ${Expr(expr)}';
            case AFunctionCall("$array_length", [AName(a)]): Expr(AFieldAccess(a, "size()"));
            case AFunctionCall("Float", [expr]): Expr(expr);
            case AFunctionCall(name, arguments):
                if ((name == ">" || name == "==" || name == "!=" || name == "+" || name == "-" || name == "*" || name == "/" || name == "and" || name == "&") && arguments.length == 2) {
                    if (name == "and") { name = "&&"; }
                    '${Expr(arguments[0])} $name ${Expr(arguments[1])}';
                } else {
                    if (name == "Int") { name = "int"; }
                    '$name(${[for (a in arguments) Expr(a)].join(", ")})';
                }
            case AMethodCall(name, method, arguments):
                if (method == "push") { method = "push_back"; }
                '$name.$method(${[for (a in arguments) Expr(a)].join(", ")})';
            case _: super.Expr(e);
        };
    }
                                                         

    function Stmt(level:Int, e:AExpr) {
        return switch (e) {
            case ABlock(exprs): '${indent(level-1)}{\n' + [for (e in exprs) Stmt(level, e)].join("") + '${indent(level-1)}}\n';
            case AFor(name, AFunctionCall("#range", [begin, AFunctionCall("+", [end, AInt(1)])]), body): '${indent(level)}for (int $name = ${Expr(begin)}; $name <= ${Expr(end)}; $name++)\n${Stmt(1+level, body)}';
            case AFor(name, AFunctionCall("#range", [begin, end]), body): '${indent(level)}for (int $name = ${Expr(begin)}; $name < ${Expr(end)}; $name++)\n${Stmt(1+level, body)}';
            case AFor(name, range, body): '${indent(level)}for (auto $name : ${Expr(range)})\n${Stmt(1+level, body)}';
            case AIf(test, ifTrue, ABlock([])): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}';
            case AIf(test, ifTrue, ifFalse): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}${indent(level)}else\n${Stmt(1+level, ifFalse)}';
            case AFunctionCall("set!", [AName(name), expr]): '${indent(level)}$name = ${Expr(expr)};\n';
            case AFunctionCall("set!", [AArrayIndex(AName(name), index), expr]): '${indent(level)}$name[${Expr(index)}] = ${Expr(expr)};\n';
            case AVar(name, type, init): '${indent(level)}${Type(type)} $name = ${Expr(init)};\n';
            case AReturn(expr): '${indent(level)}return ${Expr(expr)};\n';
            case _: '${indent(level)}${Expr(e)};\n';
        };
    }

    
    function Decl(decl:DExpr) {
        return switch(decl) {
            case DMethod(name, type, params, body): '${Type(type)} $name(${[for (p in params) Type(p.type) + " " + p.name].join(", ")})\n${Stmt(1, body)}\n\n';
            case DStaticVar(name, type, init): 'const ${Type(type)} $name = ${Expr(init)};\n';
            case _: ''; // fields handled elsewhere
        };
    }

                              
    function outputHeader(classname:String, decls:Array<DExpr>) {
        var output = "";
        var fields = collectFields(decls);
        var preconditions = collectPreconditions(decls);
        if (fields.length > 0) {
            output += 'struct ${classname}\n{\n';
            for (f in fields) {
                output += '${indent(1)}const ${Type(f.type)} ${f.name};\n';
            }
            output += '${indent(1)}${classname}(';
            output += [for (f in fields) '${Type(f.type)} ${f.name}_'].join(", ");
            output += "): ";
            output += [for (f in fields) '${f.name}(${f.name}_)'].join(", ");
            if (preconditions != null) {
                output += ' {\n${indent(2)}if (${Expr(preconditions.test)}) throw ${Expr(preconditions.message)};\n${indent(1)}}\n';
            } else  {
                output += " {}\n";
            }
            output += "};\n\n";
        }
        return output;
    }
    

    function outputDefinition(classname:String, decls:Array<DExpr>) {
        return [for (decl in decls) Decl(decl)].join("");
    }

    
    static function main() {
        var output = new OutputCpp();
        
        Sys.println('// Generated code -- CC0 -- No Rights Reserved -- http://www.redblobgames.com/grids/hexagons/

#include <cmath>
#include <cstdlib>
#include <vector>
#include <algorithm>
#include <iterator>
using std::abs;
using std::max;
using std::vector;

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
');
        
        Output.outputAll(output.outputHeader, false);

        Sys.println('// Forward declarations
');
        
        Output.outputAll(output.outputDefinition, false);

        Sys.println('
// Tests

#include <iostream>

void complain(const char* name) 
{
  std::cout << "FAIL " << name << std::endl;
}
');

        Output.outputTests(output.outputHeader);
        Output.outputTests(output.outputDefinition);
        
        Sys.println('

int main() {
  test_all();
}
');
        
    }
}

