// From http://www.redblobgames.com/grids/hexagons/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>
//
// Javascript output

using Ast;
import Output.*;

class OutputJavascript extends Output {
    function new() {}

    
    override function Type(t:TExpr) {
        return switch (t) {
            case TFractionalHex: 'Hex';
            case _: super.Type(t);
        };
    }

    
    override function Expr(e:AExpr) {
        return switch (e) {
            case AName(name):
                if (name == "PI") { name = "Math.PI"; }
                '$name';
            case ANew(TArray(_), arguments): '[${[for (a in arguments) Expr(a)].join(", ")}]';
            case ANew(type, arguments): '${Type(type)}(${[for (a in arguments) Expr(a)].join(", ")})';
            case AFunctionCall("#negate", [expr]): '-${Expr(expr)}';
            case AFunctionCall("#not", [expr]): '!${Expr(expr)}';
            case AFunctionCall("$error", [expr]): 'throw ${Expr(expr)}';
            case AFunctionCall("$array_length", [AName(a)]): Expr(AFieldAccess(a, "length"));
            case AFunctionCall("Float", [expr]): Expr(expr);
            case AFunctionCall("Int", [expr]):
                /* NOTE: it turns out that the only places I use Int,
                 * it's already an integer (either the result of
                 * Math.round or it's dividing an even number by two),
                 * so it just needs a type cast. In Javascript there's 
                 * no separate int type so nothing needed here. */
                '${Expr(expr)}';
            case AFunctionCall(name, arguments):
                if ((name == ">" || name == "==" || name == "!=" || name == "+" || name == "-" || name == "*" || name == "/" || name == "and" || name == "&") && arguments.length == 2) {
                    if (name == "and") { name = "&&"; }
                    if (name == "==") { name = "==="; }
                    if (name == "!=") { name = "!=="; }
                    '${Expr(arguments[0])} $name ${Expr(arguments[1])}';
                } else {
                    if (name == "sin") { name = "Math.sin"; }
                    if (name == "cos") { name = "Math.cos"; }
                    if (name == "round") { name = "Math.round"; }
                    if (name == "abs") { name = "Math.abs"; }
                    if (name == "sqrt") { name = "Math.sqrt"; }
                    if (name == "max") { name = "Math.max"; }
                    '$name(${[for (a in arguments) Expr(a)].join(", ")})';
                }
            case AMethodCall(name, method, arguments):
                '$name.$method(${[for (a in arguments) Expr(a)].join(", ")})';
            case _: super.Expr(e);
        };
    }


    function Stmt(level:Int, e:AExpr) {
        return switch (e) {
            case ABlock(exprs): '${indent(level-1)}{\n' + [for (e in exprs) Stmt(level, e)].join("") + '${indent(level-1)}}\n';
            case AFor(name, AFunctionCall("#range", [begin, AFunctionCall("+", [end, AInt(1)])]), body): '${indent(level)}for (var $name = ${Expr(begin)}; $name <= ${Expr(end)}; $name++)\n${Stmt(1+level, body)}';
            case AFor(name, AFunctionCall("#range", [begin, end]), body): '${indent(level)}for (var $name = ${Expr(begin)}; $name < ${Expr(end)}; $name++)\n${Stmt(1+level, body)}';
            case AFor(name, range, body): '${indent(level)}for (Object $name : ${Expr(range)})\n${Stmt(1+level, body)}';
            case AIf(test, ifTrue, ABlock([])): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}';
            case AIf(test, ifTrue, ifFalse): '${indent(level)}if (${Expr(test)})\n${Stmt(1+level, ifTrue)}${indent(level)}else\n${Stmt(1+level, ifFalse)}';
            case AFunctionCall("set!", [AName(name), expr]): '${indent(level)}$name = ${Expr(expr)};\n';
            case AFunctionCall("set!", [AArrayIndex(AName(name), index), expr]): '${indent(level)}$name[${Expr(index)}] = ${Expr(expr)};\n';
            case AVar(name, type, init): '${indent(level)}var $name = ${Expr(init)};\n';
            case AReturn(expr): '${indent(level)}return ${Expr(expr)};\n';
            case _: '${indent(level)}${Expr(e)};\n';
        };
    }

    
    // Statements (block, for, if, var) will return spaces + ... + "\n"
    // Decls (fn, static var)
    function Decl(decl:DExpr) {
        return switch(decl) {
            case DMethod(name, type, params, body): 'function $name(${[for (p in params) p.name].join(", ")})\n${Stmt(1, body)}\n';
            case DStaticVar(name, type, init): 'var $name = ${Expr(init)};\n';
            case DField(name, type): ''; // These are handled elsewhere
            case DPrecondition(_, _): ''; // These are handled elsewhere
            case _: '#unknown';
        };
    }

    
    function output(classname:String, decls:Array<DExpr>) {
        var output = [for (decl in decls) Decl(decl)].join("");
        var fields = collectFields(decls);
        var preconditions = collectPreconditions(decls);
        var preconditionTest = "";
        if (preconditions != null) {
            preconditionTest = 'if (${Expr(preconditions.test)}) throw ${Expr(preconditions.message)};\n    ';
        }
        if (fields.length > 0) {
            var fieldNames:Array<String> = [for (f in fields) f.name];
            var init:String = [for (f in fieldNames) '$f: $f'].join(", ");
            output = '\n\nfunction ${classname}(${fieldNames.join(", ")}) {\n    ${preconditionTest}return {${init}};\n}\n\n' + output;
        }
        return output;
    }


    function outputExports(className:String, decls:Array<DExpr>) {
        var names = [className];
        for (decl in decls) {
            switch (decl) {
                case DMethod(name, _, _, _): names.push(name);
                case DStaticVar(name, _, _): names.push(name);
                case _:
            }
        }
        return [for (name in names) 'exports.${name} = ${name};\n'].join("");
    }
    
                
    public static function main() {
        var output = new OutputJavascript();
        
        Sys.println('// Generated code -- CC0 -- No Rights Reserved -- http://www.redblobgames.com/grids/hexagons/
"use strict";
');
        
        Output.outputAll(output.output, true);

        Sys.println('

// Tests

function complain(name) {
    console.log("FAIL", name);
}
');
            
        Output.outputTests(output.output);

        Sys.println('
test_all()

// Exports for node/browserify modules:
');

        Output.outputAll(output.outputExports, true);
    }
}

    
