// From http://www.redblobgames.com/grids/hexagons/
// Copyright 2015 Red Blob Games <redblobgames@gmail.com>
// License: Apache v2.0 <http://www.apache.org/licenses/LICENSE-2.0.html>
//
// Python 3 output

using Ast;
import Output.*;

class OutputPython extends Output {
    function new() {}

    
    override function Type(t:TExpr) {
        return switch (t) {
            case TInt: 'int';
            case TFloat: 'float';
            case TFractionalHex: 'Hex';
            case _: super.Type(t);
        };
    }

    
    override function Expr(e:AExpr) {
        return switch (e) {
            case AName(name):
                if (name == "PI") { name = "math.pi"; }
                '$name';
            case ANew(TArray(_), arguments): '[${[for (a in arguments) Expr(a)].join(", ")}]';
            case ANew(type, arguments): '${Type(type)}(${[for (a in arguments) Expr(a)].join(", ")})';
            case AFunctionCall("$error", [expr]): 'raise ValueError(${Expr(expr)})';
            case AFunctionCall("#negate", [expr]): '-${Expr(expr)}';
            case AFunctionCall("#not", [expr]): 'not ${Expr(expr)}';
            case AFunctionCall("Int", [AFunctionCall("/", [a, b])]): '${Expr(a)} // ${Expr(b)}';
            case AFunctionCall("Float", [expr]): Expr(expr);
            case AFunctionCall(name, arguments):
                if ((name == ">" || name == "==" || name == "!=" || name == "+" || name == "-" || name == "*" || name == "/" || name == "and" || name == "&") && arguments.length == 2) {
                    '${Expr(arguments[0])} $name ${Expr(arguments[1])}';
                } else {
                    if (name == "Int") { name = "int"; }
                    if (name == "#range") { name = "range"; }
                    if (name == "$array_length") { name = "len"; }
                    if (name == "sqrt") { name = "math.sqrt"; }
                    if (name == "cos") { name = "math.cos"; }
                    if (name == "sin") { name = "math.sin"; }
                    '$name(${[for (a in arguments) Expr(a)].join(", ")})';
                }
            case AMethodCall(name, method, arguments):
                if (method == "push") { method = "append"; }
                '$name.$method(${[for (a in arguments) Expr(a)].join(", ")})';
            case _: super.Expr(e);
        };
    }


    function Stmt(level:Int, e:AExpr) {
        return switch (e) {
            case ABlock([]): '${indent(level)}pass\n';
            case ABlock(exprs): [for (e in exprs) Stmt(level, e)].join("");
            case AFor(name, range, body): '${indent(level)}for $name in ${Expr(range)}:\n${Stmt(1+level, body)}';
            case AIf(test, ifTrue, ABlock([])): '${indent(level)}if ${Expr(test)}:\n${Stmt(1+level, ifTrue)}';
            case AIf(test, ifTrue, ifFalse): '${indent(level)}if ${Expr(test)}:\n${Stmt(1+level, ifTrue)}${indent(level)}else:\n${Stmt(1+level, ifFalse)}';
            case AFunctionCall("set!", [AName(name), expr]): '${indent(level)}$name = ${Expr(expr)}\n';
            case AFunctionCall("set!", [AArrayIndex(AName(name), index), expr]): '${indent(level)}$name[${Expr(index)}] = ${Expr(expr)}\n';
            case AVar(name, _, init): '${indent(level)}$name = ${Expr(init)}\n';
            case AReturn(expr): '${indent(level)}return ${Expr(expr)}\n';
            case _: '${indent(level)}${Expr(e)}\n';
        };
    }

    
    // Statements (block, for, if, var) will return spaces + ... + "\n"
    // Decls (fn, static var)
    function Decl(decl:DExpr) {
        return switch(decl) {
            case DMethod(name, type, params, body): 'def $name(${[for (p in params) p.name].join(", ")}):\n${Stmt(1, body)}\n';
            case DStaticVar(name, type, init): '$name = ${Expr(init)}\n';
            case DField(name, type): ''; // These are handled elsewhere
            case DPrecondition(_, _): ''; // These are handled elsewhere
            case _: '#unknown';
        };
    }

    
    function output(classname:String, decls:Array<DExpr>) {
        var output = [for (decl in decls) Decl(decl)].join("");
        var fields = collectFields(decls);
        if (fields.length > 0) {
            var classDef = "";
            var preconditions = collectPreconditions(decls);
            var fieldNames:Array<String> = [for (f in fields) f.name];
            var formatted:String = [for (f in fieldNames) '"${f}"'].join(", ");
            classDef = '${classname} = collections.namedtuple("${classname}", [${formatted}])\n';
            if (preconditions != null) {
                classDef = '_${classDef}def ${classname}(${fieldNames.join(", ")}):\n${indent(1)}assert not (${Expr(preconditions.test)}), ${Expr(preconditions.message)}\n${indent(1)}return _${classname}(${fieldNames.join(", ")})\n';
            }    
            output = '\n\n${classDef}\n${output}';
        }
        return output;
    }

    
    public static function main() {
        var output = new OutputPython();
        
        Sys.println('# Generated code -- CC0 -- No Rights Reserved -- http://www.redblobgames.com/grids/hexagons/

from __future__ import division
from __future__ import print_function
import collections
import math
');
        
        Output.outputAll(output.output, true);

        Sys.println('

# Tests

def complain(name):
    print("FAIL {0}".format(name))
');
            
        Output.outputTests(output.output);

        Sys.println("
if __name__ == '__main__': test_all()
");
        
    }
}
